#ifndef _COORDS_H
#define _COORDS_H
#include <iostream>
#include <iomanip>
#include <sstream>
#include <cmath>
#include <algorithm>
#include <stdexcept>
#include <typeinfo>
#include <astrotime.h>
#include <infratypes.h>

template<class T> T sexagesimal( const std::string& cstr ) {
    std::string::size_type i = 0;
    std::string str(cstr);
    std::replace( str.begin(), str.end(), ':', ' ' );
    if( ( i = str.find_first_not_of( " " ) ) == std::string::npos ) return 0;
    double sgn = ( str.at(i) == '-' )? -1.0/3600 : 1.0/3600;
    double h2, h1, h0;
    std::istringstream istr(str);
    istr >> h2 >> h1 >> h0;
    return  sgn*( fabs(h2)*3600 + h1*60 + h0 );
}

struct Hour {
    double h;
    Hour(double x = 0) : h(x) {}
    Hour(const std::string& str) : h( sexagesimal<double>(str)/R ) {}
    operator double () const { return h; }
    static const double R = 12.0/M_PI;
};

struct Degree {
    double d;
    Degree(double x=0) : d(x) {}
    Degree(const std::string& str) : d( sexagesimal<double>(str)/R ) {}
    operator double () const { return d; }
    static const double R = 180.0/M_PI;
};

namespace std{
    std::ostream& operator<< ( std::ostream&, const Hour& );
    std::ostream& operator<< ( std::ostream&, const Degree& );
    std::istream& operator>> ( std::istream&, Hour& );
    std::istream& operator>> ( std::istream&, Degree& );
    std::istream& operator>> ( std::istream&, Date& );
    std::istream& operator>> ( std::istream&, Time& );
    std::ostream& operator<< ( std::ostream&, const Date& );
    std::ostream& operator<< ( std::ostream&, const Time& );
    std::ostream& sdeg( std::ostream& stm );
    std::ostream& udeg( std::ostream& stm );
    std::ostream& colon( std::ostream& stm );
}

class Geo {
    double m_x;
    double m_y;
public:
    static AstroDateTime utc;
    double longitude() const { return m_x; }
    double latitude()  const { return m_y; }
    double st() const { return utc.rsidTime( m_x ); }
    Geo( double lg, double lat ) : m_x(lg), m_y(lat) {}
};

double airmass( double, double, double );

class hor {
public:
    static double to_ra( double x, double y, const Geo* site ) {
        return nint( site->st() - atan2( cos(y)*sin(x), cos(site->latitude())*sin(y) + sin(site->latitude())*cos(y)*cos(x) ) );
    }
    static double to_de( double x, double y, const Geo* site ) {
        return asin( sin(site->latitude())*sin(y) - cos(site->latitude())*cos(y)*cos(x) );
    }
    static double fr_ra( double x, double y, const Geo* site  ) {
        double t =  site->st() - x;
        return atan2( cos(y)*sin(t), -cos(site->latitude())*sin(y) + sin(site->latitude())*cos(y)*cos(t) );
    }
    static double fr_de( double x, double y, const Geo* site  ) {
        double t = site->st() - x;
        return asin( sin(site->latitude())*sin(y) + cos(site->latitude())*cos(y)*cos(t) );
    }
};

class equ {
public:
    static double to_ra( double x, double y, const Geo* site  ) { return x; }
    static double to_de( double x, double y, const Geo* site  ) { return y; }
    static double fr_ra( double x, double y, const Geo* site  ) { return x; }
    static double fr_de( double x, double y, const Geo* site  ) { return y; }
};

class eqt {
public:
    static double to_ra( double x, double y, const Geo* site ) {
        return nint( site->st() - x );
    }
    static double to_de( double x, double y, const Geo* site ) { return y; }
    static double fr_ra( double x, double y, const Geo* site ) {
        return sint( site->st() - x );
    }
    static double fr_de( double x, double y, const Geo* site ) { return y; }
};

class ecl {
   static const double ce = 0.917475; // cos(23d26'25")
   static const double se = 0.397793; // sin(23d26'25")
public:
   static double to_ra( double x, double y, const Geo* site ) {
       return nint( atan2(sin(x)*cos(y)*ce - sin(y)*se, cos(x)*cos(y)) );
   }
   static double to_de( double x, double y, const Geo* site ) {
       return asin( sin(y)*ce + sin(x)*cos(y)*se );
   }
   static double fr_ra( double x, double y, const Geo* site ) {
       return sint( atan2(sin(x)*cos(y)*ce + sin(y)*se, cos(x)*cos(y)) );
   }
   static double fr_de( double x, double y, const Geo* site ) {
       return asin( sin(y)*ce - sin(x)*cos(y)*se );
   }
};

template <class T> class Coord {
    double m_x;
    double m_y;
    const Geo* p_site;
    friend class hor;
public:
    double x() const { return m_x; }
    double y() const { return m_y; }
    const Geo* site() const { return p_site; }
    Degree  az( double t=0 ) const {
        return (typeid(T) == typeid(hor) || !ptrtst() ) ? x() : hor::fr_ra( T::to_ra( m_x, m_y, p_site ), T::to_de( m_x, m_y, p_site ), p_site );
    }
    Degree alt( double t=0 ) const {
        return (typeid(T) == typeid(hor) || !ptrtst() ) ? y() : hor::fr_de( T::to_ra( m_x, m_y , p_site ), T::to_de( m_x, m_y, p_site ), p_site  );
    }
    Hour    ra( double t=0 ) const {
        return (typeid(T) == typeid(equ) || !ptrtst() ) ? x() : T::to_ra( m_x, m_y, p_site  );
    }
    Degree dec( double t=0 ) const {
        return (typeid(T) == typeid(equ) || !ptrtst() ) ? y() : T::to_de( m_x, m_y, p_site  );
    }
    Hour    ha( double t=0 ) const {
        return (typeid(T) == typeid(eqt) || !ptrtst() ) ? x() : eqt::fr_ra( T::to_ra( m_x, m_y, p_site  ), T::to_de( m_x, m_y, p_site  ), p_site );
    }
    Coord( const double ix=0, const double iy=0,  const Geo* s = 0  ) : m_x(ix), m_y(iy), p_site(s) {}
    Coord& operator=(const Coord& z ) {m_x = z.x(); m_y = z.y(); p_site = ( z.site() == 0 )? p_site : z.site(); return *this;}
    template<class U>
    Coord( const Coord<U>& z ) : p_site(z.site()) {
        ptrtst();
        double v = U::to_ra( z.x(), z.y(), z.site() );
        double w = U::to_de( z.x(), z.y(), z.site() );
        m_x = T::fr_ra( v, w, z.site() );
        m_y = T::fr_de( v, w, z.site() );
    }
private:
    bool ptrtst() const { return ( p_site == 0 )? throw std::runtime_error("Geo undefined" ) : true; }
public:
    point<double> rate() {
        double cl = cos( p_site->latitude() );
        double sl = sin( p_site->latitude() );
        return point<double>( sl + cl*tan( alt() )*cos( az() ), -cl*sin( az() ) );
    }
    point<double> skew() {
        double cl = cos( p_site->latitude() );
        double sl = sin( p_site->latitude() );
        double t = p_site->st() - ra();
        return point<double>( (sl*cos(dec()) - cl*sin(dec())*cos(t))/cos(alt()), cl*sin(t)/cos(alt()) );
    }
    double distance( const Coord& b ) {
        return acos( sin(m_y)*sin(b.y()) + cos(m_y)*cos(b.y())*cos(m_x-b.x()));
    }
    double Mz() {
        return 1.0/sin(alt());
    }
    double Mz( const Date& da, const Time& ut ) {
        const double mu = 1.0027379093;
        double ha = ( s0( (int)da ) + mu*ut/24.0 )*M_PI*2 + p_site->longitude() - m_x;
        return airmass( ha, m_y, p_site->latitude() );
    }
    double star_set( double alt ) {
        double c = ( sin(alt) - sin(p_site->latitude())*sin( dec()))/cos(p_site->latitude())/cos( dec());
        return ( std::abs(c) > 1.0 ) ? 0.0 : acos( c );
    }
};

#endif
