#include <cerrno>
#include <cstring>
#include <cstdlib>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <sstream>

#include <astrotime.h>
#include <errlog.h>
#include "../defnames.h"

AstroDateTime utc_err;

namespace {
    std::ostream& operator<<( std::ostream& stm, const std::ostringstream& str ){
        std::cerr << str.str() << std::endl;
        stm << str.str() << std::endl;
        return stm;
    }
}

namespace { 
    std::string timestamp() {
        utc_err.current(CONV);
        std::ostringstream comstr;
        comstr << utc_err.strDate() << ' ' << utc_err.strTime(2) << ' ';
        return comstr.str();
    }
}

Error::Error( const ErrorsCode c,  const std::string& par )
      : std::runtime_error(par), m_cod(c) { 
          utc_err.current(CONV);
      }

Errlog::Errlog( ) {
    open( utc_err.strFileName() );
}

Errlog::~Errlog() {
}

void Errlog::open( const std::string name ) {
    std::string filename = program_log + name + extension_log;
    current_log = name;
    log.open( filename.c_str(), std::ios::out | std::ios::app );
    if( !log ) write( (ErrorsCode)errno, filename );
    log << timestamp() << progname << VERSION << " Compiled " << __TIME__ << " " << __DATE__ << std::endl;
}

void Errlog::reopen() {
    if( utc_err.strFileName() == current_log) return;
    log.close();
    open(utc_err.strFileName());
}

void Errlog::write( const Error& s ) {
    write( s.err(), s.what() );
}

void Errlog::write( const std::exception& s ) {
    write( ERR_STD_EXCP, s.what() );
}

void Errlog::write( ErrorsCode en, const std::string& s ) {
    std::ostringstream errstr;
    errstr << timestamp() << "(" << std::setw(3) << std::setfill('0')
           << en << ") " << s << " - " << strerror(en);
    log << errstr;
    last_error = errstr.str();
}

void Errlog::write( const std::string& s ) {
    std::ostringstream errstr( timestamp() + "(000) " + s);
    log << errstr;
}

void Errlog::write_if_test( const std::string& s ) {
    if( testmode && s.size() ) {
        std::ostringstream errstr( timestamp() + "(700) " + s);
        log << errstr;
    }
}

void Errlog::wthrow( ErrorsCode en, const std::string& s ) {
    if( !testmode ) throw ErrFat( en, s );
    ErrlogSingl::instance().write( en, s );
}

void Errlog::warning( const std::string &s, double a  ) {
    std::ostringstream mess;
    mess << s << " > " << std::setprecision(2) << std::abs(a);
    write( mess.str() );
}

void Errlog::warning( const std::string &s, int rp  ) {
    std::ostringstream mess;
    mess << s << " [" << rp << "]"; 
    write( mess.str() );
}

void Errlog::warning( const std::string &s ) {
    if( last_warning == s ) {
        repeated++;
        return;
    }
    if( repeated ) {
        warning( last_warning, repeated );
        repeated = 0;
    }
    last_warning = s;
    write( s );
}

const int ERSBASE = 300;
const int TURBINA_ERRBASE = 600;

const char* Errlog::strerror(  ErrorsCode err ) {
    if( err < ERSBASE ) return ::strerror( (int)err);
//    if( err < TURBINA_ERRBASE ) return rs_geterror(err);
    switch(err) {
        case    ERR_CFG_STRUCT      :  return "Invalid structure";
        case    ERR_CFG_MISSED      :  return "Missed in configuration";
        case    ERR_INI_OBJECT      :  return "Is not been created";
        case    ERR_CENT_EMPTY      :  return "Centering frame is empty";
        case    ERR_CCD_EMPTY       :  return "CCD frame is empty";
        case    ERR_NORM_EMPTY      :  return "Subframe is empty";
        case    ERR_NORM_SHIF       :  return "Center mode wrong data";
        case    ERR_PROT_SYNT       :  return "Protocol syntax error";
        case    ERR_PROT_WRCOMM     :  return "Protocol wrong command";
        case    ERR_PROT_NOKWD      :  return "Protocol keyword missed";
        case    ERR_PROT_WRKWD      :  return "Protocol wrong keyword";
        case    ERR_PROT_NOVAL      :  return "This keyword(s) requires value";
        case    ERR_PROT_SCEN       :  return "Invalid symbol in scenario";
        case    ERR_REG_STRUCT      :  return "Invalid regular expression";
        case    ERR_IEEE_BUS        :  return "IEEE1394 bus";
        case    ERR_CAM_INIT        :  return "IIDC camera initialization";
        case    ERR_CAM_NOTF        :  return "No IIDC camera found";
        case    ERR_CAM_IDEN        :  return "No specified camera found";
        case    ERR_CAM_ERR         :  return "IIDC camera setting";
        case    ERR_CFG_CHANNEL     :  return "Invalid channels configuration";
        case    ERR_NOT_CONFIGURED  :  return "Module is not configured";
        case    ERR_UNKNOWN_ADDRESS :  return "Unknown address";
        case    ERR_DRIVER_VERSION  :  return "Unsupported driver version";
        case    ERR_NOTHING_DONE    :  return "No measurement done yet"; 
        case    ERR_EXCHANGE        :  return "RS485 error";
        case    ERR_EXC_STAT        :  return "RS485 status error";
        case    ERR_NOT_REPLY       :  return "Module doesn't respond";
        case    ERR_SOCK_TOUT       :  return "Network communication";
        case    ERR_SOCK_BAD        :  return "Client detached";
        case    ERR_IDENT           :  return "Module has wrong identification";
        case    ERR_TEST            :  return "Test failed";
//        case    ERR_OVRLIGHT        :  return "Overlight! HV is off";
//        case    ERR_MIRROR          :  return "Mirror on Axis";
        case    ERR_DEV_FUNC        :  return "Device function error";
        case    ERR_DEV_STAT        :  return "Device status error";
        case    ERR_DEV_RANGE       :  return "Device range error";
        case    ERR_ALG             :  return "Algorithmic Error";
        case    ERR_EMPTY_MODE      :  return "Empty scenario run";
        case    ERR_CMD_OPTION      :  return "Illegal options"; 
        case    ERR_STD_EXCP        :  return "Std::exception";
        case    ERR_SERIAL          :  return "Serial communication";
        case    ERR_PROT_MEADE      :  return "Meade protocol error";
        case    WARNING             :  return "This is warning only";
        case    LOG_ONLY            :  return "Write in the log only";
        case    ERR_UKN_EXCP        :  return "Unknown type exception";
    }
    return "Unknown error";
}
