#include <sys/ioctl.h>
#include <fcntl.h>
#include <sstream>
#include <cerrno>
#include <cstring>
#include <sys/time.h>
#include <vector>
#include <map>
#include <iostream>
#include <iomanip>
#include <signal.h>
#include <cmath>

#include <mcua.h>
#include <rserrno.h>
#include <exchange.h>

#include <unistd.h> // write, usleep, close

const int MAXRETRY  = 2; // Maximal number of transaction attempts
const int MAX_PACKET_SIZE = 32; // Maximal packet size

const int ACK = SignalACK;
const int NAK = SignalNAK;
const int NOD = SignalNOD;
const int ACN = SignalACN;
const int ACY = SignalACY;
const int ACW = SignalACW;
const int SINC= SignalSINC;
const int DNG = SignalDNG;
const int PKL = SignalPKL;
const int PBO = SignalPBO;
const int DSC = SignalDSC;
const int COL = SignalCOL;
const int EER = SignalEER;
const int UND = SignalUND;

enum { INFO_SIG, LOOP_SIG };

struct Signl {
    int stype;
    std::string abbr;
    std::string text;
    Signl(const int st, const std::string& a, const std::string& t)
          : stype(st), abbr(a), text(t) {};
    Signl() {};
};

namespace Signal {
    const std::string& abbrev( int code);
    const std::string& message( int code);
    int type( int code);
    typedef std::pair< int, struct Signl> PAIR;
    PAIR p[] = {
        PAIR( ACK, Signl(INFO_SIG, "ACK", "Successful data transmission")),
        PAIR( NOD, Signl(INFO_SIG, "NOD", "Required data are not ready")),
        PAIR( ACN, Signl(INFO_SIG, "ACN", "This command does not exist")),
        PAIR( ACY, Signl(INFO_SIG, "ACY", "Acknowledgement of existing command")),
        PAIR( ACW, Signl(INFO_SIG, "ACW", "Command can't be executed now")),
        PAIR( DSC, Signl(LOOP_SIG, "DSC", "Module doesn't respond")),
        PAIR( NAK, Signl(LOOP_SIG, "NAK", "Broken packet was received")),
        PAIR( COL, Signl(LOOP_SIG, "COL", "Transmission error or packets collision")),
        PAIR( UND, Signl(LOOP_SIG, "UND", "Unexpected data instead of signal ")),
        PAIR( EER, Signl(LOOP_SIG, "EER", "General exchange error"))
    };
    std::map< int, Signl> sg( p, p + sizeof(p)/sizeof(p[0]));
}

int Signal::type( int code) {
    return (sg.find(code) == sg.end())? sg[EER].stype : sg[code].stype;
}

const std::string& Signal::abbrev( int code) {
    return (sg.find(code) == sg.end())? sg[EER].abbr : sg[code].abbr;
}

const std::string& Signal::message( int code) {
    return (sg.find(code) == sg.end())? sg[EER].text : sg[code].text;
}

namespace {
    const std::string RS_strerror(int errnum) {
        switch(errnum) {
            case ERSNODEVICE    :
            case ENOENT         : return "No RS485 device";
            case ERSNOPORT      : return "No this port";
            case ERSNOINTERFACE : return "No RS485 interface";
            case ENODEV         :
            case ERSNODRIVER    : return "No RS485 driver";
            case ERSBUSY        : return "RS485 device is busy";
            case ERSSHUTDOWN    : return "Device is off or disconnected";
            case ERSISATMOUT    : return "RS485 ISA time out";
            case ERSNOCONVREPLY : return "No reply from RS485 converter";
            case ERSFRAME       : return "RS485 frame error";
            case ERSFIFOVERFLOW : return "RS485 FIFO overflow";
            case ERSNOIOCTL     : return "RS485: no such ioctl command";
            default             : if( errnum < ERSBASE ) return std::strerror(errnum);
                                  return "Unknown RS485 error";
        }
    }
}

/*
    int micro_delay( int mks ) {
        struct timeval moms;
        struct timeval mome;
        int mksd = 0;
        int secd = 0;
        gettimeofday( &moms, NULL );
        while( mksd < mks ) {
            gettimeofday( &mome, NULL );
            secd = mome.tv_sec - moms.tv_sec;
            mksd = mome.tv_usec - moms.tv_usec;
            if( secd < 0 || secd > 1 ) return( -1);
            if( secd > 0 ) mksd += 1000*1000;
        }
        return( 0 );
    }

    int micro_wait( int mks ) {
        static struct timeval moms;
        struct timeval mome;
        int mksd = 0;
        int secd = 0;
        if( mks == 0 ){
            gettimeofday( &moms, NULL );
            return( 0 );
        }
        while( mksd < mks ) {
            gettimeofday( &mome, NULL );
            secd = mome.tv_sec - moms.tv_sec;
            mksd = mome.tv_usec - moms.tv_usec;
            if( secd < 0 || secd > 1 ) return( -1);
            if( secd > 0 ) mksd += 1000*1000;
        }
        return( 0 );
    }

*/

RS485impl::RS485log::RS485log( const unsigned char *inp,  const int nbyte, RS485impl* ptr ) : ptr_up(ptr) {
        s << std::hex << std::fixed << std::setfill('0');
        for( int i=0; i < nbyte; i++ ) s << std::setw(2) << (int)inp[i] << " ";
        s << "<- ";
}

RS485impl::RS485log::~RS485log() {
    ptr_up->logstr = "-> " + s.str();
}

void RS485impl::RS485log::log( const std::string& t ) {
    s << t << " ";
}

void RS485impl::RS485log::log( unsigned char *out, const int n ) {
        for( int i=0; i < n; i++ ) s << std::setw(2) << (int)out[i] << " ";
}

int RS485impl::transaction(const unsigned char *buffi, const int nbyte, unsigned char *buffo) {
    int reply = 0;
    char temporal[IO_BUFF_SIZE];
    int attempt = MAXRETRY;
    packet_number++;
    if( nbyte > MAX_PACKET_SIZE ) throw ErrFatal( "Too long command. Why??" );
    RS485log scr( buffi, nbyte, this );
    RSTimer time = RSTimer( 1.0 ); // 0.5
/*    if( packet_number%17 == 0 ) {           // RS Timer debug
        timespec a;
        a.tv_sec  = (long)2;
        a.tv_nsec = (long)rint(0.5e+9);
        if( nanosleep( &a, NULL ) < 0 ) std::cerr << "Nanosleep - Action was interrupted" << std::endl;
        scr.log( Signal::abbrev(EER));
        return EER;
    }*/
    while( attempt-- ) {
        memcpy( temporal, buffi, nbyte );
        reply = write( id_mcua, temporal, nbyte );
        if( time.out() ) throw ErrDriver( "RS driver timeout" );
        if( reply < 0 ) {
            errors_number++;
            scr.log(RS_strerror(errno));
            if( attempt == 0 ) {
                throw ErrDriver( RS_strerror(errno) );
            }
            usleep(10000); //   RESET
            ioctl( id_mcua, RS_RESET );
            scr.log( " RESET " );
            continue;
        }
        if( reply <= MAX_PACKET_SIZE) {         // data, no Signal
            memcpy( buffo, temporal+1, --reply );
            scr.log( buffo, reply );
            return reply;
        }
        if( Signal::type(reply) == INFO_SIG ) { // info Signal
            scr.log( Signal::abbrev(reply));
            return reply;
        }
        errors_number++;
        scr.log( Signal::abbrev(reply));
        usleep(1000);
    }                                           // loop Signal
    return reply;
}

RS485impl::RS485impl( const std::string& d, int r ) : packet_number(0), errors_number(0), extranum(0), device(d) {
    char vers[32];
    if( (id_mcua = open( d.c_str(), O_RDWR )) < 0 ) {
        throw ErrFatal( RS_strerror(errno) );
    }
    if( ioctl( id_mcua, RS_GET_VERS, vers )  < 0 ) {
        close( id_mcua );
        throw ErrFatal( RS_strerror(errno) );
    }
    if( r != 0 && ioctl( id_mcua, RS_SET_BAUDRATE, r ) < 0) {
        close( id_mcua );
        throw ErrFatal( "Unsupported exchange rate" );
    }
    logstr = std::string("Driver version ") + std::string(vers);
}

RS485impl::~RS485impl(void) {
    close( id_mcua );
}

int RS485impl::askByte( unsigned int module, const AByte command ) {
    int nsent  = 0;
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;
    int ret = transaction(strcomm, nsent, strread);
    if( ret != 1 ) throw ErrSignal( Signal::abbrev(ret) );
    return (int)(*strread);
}

int RS485impl::askWord( unsigned int module, const AWord command ) {
    int nsent = 0;
    int avalue = 0;
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;
    int ret = transaction(strcomm, nsent, strread);
    if( ret != 2 ) throw ErrSignal( Signal::abbrev(ret) );
    memcpy( (char*)&avalue, strread, 2 );
    return avalue ;
}

int RS485impl::askLong( unsigned int module, const ALong command ) {
    int nsent = 0;
    int avalue = 0;
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;
    int ret = transaction( strcomm, nsent, strread );
    if( ret != 3 ) throw ErrSignal( Signal::abbrev(ret) );
    strread[3] = ((char)strread[2] < 0 ? 0xff : 0 );
    memcpy( (char*)&avalue, strread, 4 );
    return avalue;
}

std::vector<unsigned char> RS485impl::askData( unsigned int module, const AData command ) {
    int nsent = 0;
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;
    int ret = transaction( strcomm, nsent, strread );
    if( ret < MAX_PACKET_SIZE ) return std::vector<unsigned char>( strread, strread+ret );
    if( ret != NOD ) throw ErrSignal( Signal::abbrev(ret) );
    return std::vector<unsigned char>(); //   -   
}

BaseRS485::com_stat RS485impl::sendSimpleCommand( unsigned int module, const SComm command) {
    int nsent = 0;
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;
    int ret = transaction( strcomm, nsent, strread );
    if( (ret == ACY) || (ret == ACK) ) return DONE;
    if( ret != ACW ) {
            throw ErrSignal( Signal::abbrev(ret) );
    }
    return BUSY;
}

BaseRS485::com_stat RS485impl::sendByteCommand( unsigned int module, const SByte command, char arg) {
    int nsent = 0;
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;
    strcomm[nsent++] = arg;
    int ret = transaction( strcomm, nsent, strread );
    if( (ret == ACY) || (ret == ACK) ) return DONE;
    if( ret != ACW) throw ErrSignal( Signal::abbrev(ret) );
    return BUSY;
}

BaseRS485::com_stat RS485impl::sendWordCommand( unsigned int module, const SWord command, short arg) {
    int nsent = 0;
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)command;
    strcomm[nsent++] = low( arg );
    strcomm[nsent++] = high( arg );
    int ret = transaction( strcomm, nsent, strread );
    if( (ret == ACY) || (ret == ACK) ) return DONE;
    if(ret != ACW) throw ErrSignal( Signal::abbrev(ret) );
    return  BUSY;
}

BaseRS485::com_stat RS485impl::sendData( unsigned int module, const SData length, const std::vector<unsigned char>& data) {
    int nsent = 0;
    strcomm[nsent++] = (char)module;
    strcomm[nsent++] = (char)length;
    memcpy(strcomm+2, &data[0], length );
    nsent = length+2;
    int ret = transaction( strcomm, nsent, strread );
    if( (ret == ACY) || (ret == ACK) ) return DONE;
    if(ret != ACW) throw ErrSignal( Signal::abbrev(ret) );
    return BUSY;
}

int RS485impl::sendVoid(const unsigned char *buff, const int length, unsigned char *data) {
    int nsent = length;
    memcpy( strcomm, buff, length );
    int ret = transaction( strcomm, nsent, strread );
    if( ret <= MAX_PACKET_SIZE ) {// data
        memcpy( data, strread, ret );
    } else strcpy( (char*)data, Signal::abbrev(ret).c_str());
    return ret;
}

//  ,     data.
// ,    - ?

const int headlen = 4;

int RS485impl::readData( unsigned char *head, unsigned char *data) {
    int n_read_data = 0;
    int ret = 0;
    if( (n_read_data = read(id_mcua, strdata, IO_BUFF_SIZE )) ) {
        if(n_read_data < 0) throw ErrFatal(RS_strerror(errno));
        if(n_read_data > headlen ) { // header
            ret = n_read_data-headlen;
            memcpy( head, strdata+1, headlen );
            memcpy( data, strdata+headlen+1, n_read_data-headlen );
        }
    }
    return ret;
}

int RS485impl::readData( unsigned int& from, unsigned int& numb, unsigned short *data) {
    int n_read_data = 0;
    int ret = 0;
    unsigned short pnum;
    if( (n_read_data = read( id_mcua, strdata, IO_BUFF_SIZE )) ) {
         if(n_read_data < 0) throw ErrFatal(RS_strerror(errno));
        if( n_read_data > headlen && !((n_read_data-headlen)%2)) {
            ret = (n_read_data-headlen)/2;
            from = (unsigned int)(*(strdata+2));
            memcpy( &pnum, strdata+3, 2 );
            numb = pnum + extranum;
            if( pnum == 0xFFFF ) extranum += 0x10000;
            memcpy( data, strdata+headlen+1, n_read_data-headlen );
        }
    }
    return ret;
}

void RS485impl::resetData( void ) {
    extranum = 0;
    if((ioctl( id_mcua, RS_FIFO_FLUSH ) < 0) || (ioctl(id_mcua, RS_BLNUMBER_RESET) < 0))
        throw ErrFatal( RS_strerror(errno) );
}

bool RSTimer::expired = false;

void RSTimer::catch_alarm(int sig) {
    expired = true;
}

RSTimer::RSTimer( double in ){
    expired = false;
    end.it_interval.tv_usec = 0;
    end.it_interval.tv_sec = 0;
    end.it_value.tv_usec = (unsigned int)rint( 1000000*(in - floor(in)));
    end.it_value.tv_sec = (unsigned int)floor(in);
    if( setitimer( ITIMER_REAL, &end, &old) < 0 )
        throw BaseRS485::ErrFatal( "RS timer wasn't set" );
    else  {
        if( signal( SIGALRM, catch_alarm ) == SIG_ERR ) BaseRS485::ErrFatal( "Handler wasn't set" );
    }
}

RSTimer::~RSTimer(){
    expired = false;
    end.it_interval.tv_usec = 0;
    end.it_interval.tv_sec = 0;
    end.it_value.tv_usec = 0;
    end.it_value.tv_sec = 0;
    setitimer( ITIMER_REAL, &end, NULL );
}

