#include <iostream>
#include <fstream>
#include <string>
#include <map>
#include <iomanip>
#include <time.h>
#include <math.h>

#include <module.h>
#include <errlog.h>

#include <unistd.h> // usleep

//const char LOAD_KEY   = 0x3E;
//const char READ_KEY   = 0xFE;

const SComm Module::RESET = 0x87;
const AByte Module::GET_STATUS = 0xE0;
const AByte Module::GET_SENSOR = 0xE1;
const AData Module::GET_IDENT = 0xA2;
const AData Module::GET_CONST = 0xA3;
const std::string Module::sect_device("Devices");

pthread_mutex_t Module::Lock::mtx = PTHREAD_MUTEX_INITIALIZER;
bool Module::parking = true;

namespace {
    std::istream& operator>>( std::istream& stm, unsigned char & b ){
        int ii;
        stm >> ii;
        b = (unsigned char)ii;
        return stm;
    }
    std::ostream& operator<<( std::ostream& stm, unsigned char & b ){
        int ii = (int)b;
        stm << std::hex << std::fixed << std::setfill('0') << std::setw(2) << ii;
        return stm;
    }
}

std::map< int, std::string > Module::names;

Module::Module( Config &cfg, RS485& r, const std::string& mn )
        : rs(r), mod_name(mn), mod_connected(false), status(-1), sensor(-1), Conn_thread(0) {
//    mod_verified = false;
    parking = false;
    mod_address = cfg.inp<int>( sect_device, mod_name, "Module");
    names[mod_address] = mod_name;
    ident =  cfg.inpv<unsigned char>( sect_device, mod_name, "Identification");
    try {
        verification();
        mod_verified = true;
    }
    catch ( const std::exception& e ) {
        ErrlogSingl::instance().wthrow( ERR_EXCHANGE, e.what() );
    }
}

Module::~Module() {
    parking = true;
//    std::cerr << "~Module " << mod_name << std::endl;
/*
    if(Conn_thread) {
//        std::cerr << "Conn_thread" << std::endl;
        int err;
        if( (err=pthread_cancel(this->Conn_thread)) != 0 ) {
            if( err == ESRCH ) {
                std::cerr << "Conn thread is terminated already" << std::endl;
            } else {
                std::cerr << "pthread_cancel err=" << err << std::endl;
            } 
            return;
        }    
//        std::cerr << "pthread_canceled" << std::endl;
        if( pthread_join(this->Conn_thread, NULL) != 0 ) {
            ErrlogSingl::instance().write(ERR_ALG, "Conn pthread_join failed");
        }
//        std::cerr << "pthread_joined" << std::endl;
    }
*/
//    std::cerr << "Destructor Module" << std::endl;
}

const std::string& Module::name (int addr) {
    if( names.find(addr) == names.end() ) {
        std::ostringstream s("Unknown module: ");
        s << std::hex << addr;
        throw RS485::ErrFatal( s.str() );
    }
    return( names[addr] );
}

void Module::reset() {
    sendSimpleCommand( RESET );
    usleep(2000);
}

void Module::signal_handler( RS485::ErrSignal& e ) {
    std::string signal(e.what());
    std::string status;
    if( signal == "DSC" ) status = " disconnected";
    if( signal == "ACN" ) status = " self-reset";
    if( signal == "COL" ) status = " collision";
    if( signal == "NAK" ) status = " NAK signal";
    if( signal == "EER" ) status = " exchange error";
    try {
        verification();
        ErrlogSingl::instance().write( "Module " + mod_name + " connects again" );
        return; // OK!
    }
    catch ( const std::exception& e ) {
//        std::cerr << "Ne OK! disconnected...\n";
    }
    if( mod_connected ) { //   ,     
        mod_connected = false;
        pthread_create( &Conn_thread, NULL, &Module::thread, (void*)this );
        if( status.empty() ) {
            throw RS485::ErrFatal( "Module " + mod_name + " " + signal  );
        } else {
            throw RS485::ErrFatal( "Module " + mod_name + status  );
        }
    }
    throw RS485::ErrFatal( "Module " + mod_name + " not connected"  );
//    throw std::exception();
}

const unsigned int CONN_TMOUT = 100; // sec
void *Module::thread( void* a ) {
    Module* ptr = (Module*)a;
    for( unsigned int i = 0; i < CONN_TMOUT && !parking ; i++ ) {
        try { 
            Lock lock;
            ptr->verification();
            ErrlogSingl::instance().write( "Module " + ptr->mod_name + " connects again" );
            pthread_detach( ptr->Conn_thread );
            return 0;        
        }
        catch ( RS485::ErrSignal& e ) {
//            std::cerr << "Conn_check............." << std::endl;
        }
        catch ( RS485::ErrDriver& e ) {
//            std::cerr << "Conn_check............." << std::endl;
        }
        nsleep(1); //  
    }
    pthread_detach( ptr->Conn_thread );
    return 0;
}

void Module::verification() {
    std::vector<unsigned char> real_ident = rs.askData( mod_address, GET_IDENT );
    if( real_ident.size() != 4 ) {
        usleep(1000);
        rs.sendSimpleCommand( mod_address, RESET );
        real_ident = rs.askData( mod_address, GET_IDENT );
    }
    std::string mess = mod_name + " initialization";
    if( ident != real_ident ) ErrlogSingl::instance().wthrow(ERR_IDENT, mess );
    mod_connected = true;
}

void Module::readConst() {
    constants = askData( GET_CONST );
    if( constants.size() != 4 ) {
        std::string mess = mod_name + " constants reading";
        ErrlogSingl::instance().wthrow( ERR_EXCHANGE, mess);
        constants = std::vector<unsigned char>(4,1); // for debug mode
    }
}

void Module::nsleep( double sec ) {
    timespec a;
    a.tv_sec  = (long)floor( sec );
    a.tv_nsec = (long)rint(1e+9*(sec-floor(sec)));
    if( nanosleep( &a, NULL ) < 0 ) ErrlogSingl::instance().write( "Action was interrupted" );
}

