#ifndef _RS_SKIN_H
#define _RS_SKIN_H
#include <pthread.h>
#include <cerrno>
#include <vector>
#include <memory>
#include <exchange.h>
#include <errlog.h>

class RS485 : public BaseRS485 {
    std::auto_ptr<BaseRS485> ptrs;
    std::string device;
    int baudrate;
private:
    RS485(const RS485&);
    RS485& operator=(RS485&);
    void reset();
    void error_handler( BaseRS485::ErrDriver& e );
private:
    class Lock {
        static pthread_mutex_t mtx;
        Lock(const Lock&);
        Lock& operator=(const Lock&);
        struct timespec locktime;
        struct timeval  firstime;
    public:
        Lock() {
            gettimeofday( &firstime, NULL );
            locktime.tv_sec = firstime.tv_sec + 1;
            locktime.tv_nsec = 1000*firstime.tv_usec;
            int err;
            if( (err = pthread_mutex_timedlock( &mtx, &locktime)) != 0) {
                if( err == ETIMEDOUT ) ErrlogSingl::instance().write( "RS485 Lock timeout" );
                else ErrlogSingl::instance().write( "RS485 Lock unknown error" );
            }
        }
        ~Lock() { pthread_mutex_unlock(&mtx); }
    };

    class Protocol {
        BaseRS485* ptr;
    public:
        Protocol( BaseRS485* p ) : ptr(p) {}
        ~Protocol() { ErrlogSingl::instance().write_if_test( ptr->protocol() ); }
    };

public:
    RS485(const std::string&, int baud = 0 );
    ~RS485();
    bool shadow() {return ptrs->shadow();}
    int askByte(unsigned int m_address, AByte c) try {
        Lock lock;
        Protocol p( ptrs.get() );
        return ptrs->askByte(m_address, c);
    }
    catch( ErrDriver& e) {
        error_handler( e );
        return -1;
    }
    int askWord(unsigned int m_address, AWord c) try {
        Lock lock;
        Protocol p( ptrs.get() );
        return ptrs->askWord(m_address, c);
    }
    catch( ErrDriver& e) {
        error_handler( e );
        return -1;
    }
    int askLong(unsigned int m_address, ALong c) try {
        Lock lock;
        Protocol p( ptrs.get() );
        return ptrs->askLong(m_address, c);
    }
    catch( ErrDriver& e) {
        error_handler( e );
        return -1;
    }
    std::vector<unsigned char> askData(unsigned int m_address, AData c ) try {
        Lock lock;
        Protocol p( ptrs.get() );
        return ptrs->askData(m_address, c);
    }
    catch( ErrDriver& e) {
        error_handler( e );
        return std::vector<unsigned char>();
    }
    com_stat sendSimpleCommand(unsigned int m_address, SComm c) try {
        Lock lock;
        Protocol p( ptrs.get() );
        return ptrs->sendSimpleCommand(m_address, c);
    }
    catch( ErrDriver& e) {
        error_handler( e );
        return RS485::NONE;
    }
    com_stat sendByteCommand(unsigned int m_address, SByte c, char b) try {
        Lock lock;
        Protocol p( ptrs.get() );
        return ptrs->sendByteCommand(m_address, c, b);
    }
    catch( ErrDriver& e) {
        error_handler( e );
        return RS485::NONE;
    }
    com_stat sendWordCommand(unsigned int m_address, SWord c, short w) try {
        Lock lock;
        Protocol p( ptrs.get() );
        return ptrs->sendWordCommand(m_address, c, w);
    }
    catch( ErrDriver& e) {
        error_handler( e );
        return RS485::NONE;
    }
    com_stat sendData(unsigned int m_address, const SData length, const std::vector<unsigned char> &d) try {
        Lock lock;
        Protocol p( ptrs.get() );
        return ptrs->sendData(m_address, d.size(), d);
    }
    catch( ErrDriver& e) {
        error_handler( e );
        return RS485::NONE;
    }
    void statistics() {
        std::cerr << ptrs->stat_packets() << "/" << ptrs->stat_errors() << std::endl;
    }

    int readData(unsigned char* head, unsigned char* data ) {
        Lock lock;
        return ptrs->readData( head, data );
    }

    int readData(unsigned int& addr, unsigned int& packnum,  unsigned short* data ) {
        Lock lock;
        return ptrs->readData( addr, packnum, data );
    }

    void resetData() { 
        Lock lock;
        ptrs->resetData();
    }

};

#endif
