#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <cmath>

#include  <vector>
#include  <iostream>
#include  <iomanip>
#include <stdexcept>

#include "meteo.h"

/* Wind direction measurements
    N - 0
    E - 90
    S - 180
    W - 270
The needed shift was added in 21.07.07
*/

const std::string   sect_device("Devices");

const AData Meteo::GET_PRIM_DATA = 0xA0;
const AData Meteo::GET_SECD_DATA = 0xA1;
const int SENS2 = 0x02;
const int SENS1 = 0x01;

Meteo::Meteo( Config& cfg, RS485& r, const std::string& s ): Module( cfg, r, s ) {
                                    //  
    factors.temperature1   = cfg.inp<double>( sect_device, s, "Temp1factor");
    factors.temperature2   = cfg.inp<double>( sect_device, s, "Temp2factor");
    shifts.temperature1    = cfg.inp<double>( sect_device, s, "Temp1shift");
    shifts.temperature2    = cfg.inp<double>( sect_device, s, "Temp2shift");
    factors.wind_direction = cfg.inp<double>( sect_device, s, "WindDirFactor");
    shifts.wind_direction  = cfg.inp<double>( sect_device, s, "WindDirShift");
    factors.humidity       = cfg.inp<double>( sect_device, s, "HumidityFactor");
    shifts.humidity        = cfg.inp<double>( sect_device, s, "HumidityShift");
    factors.wind_speed     = cfg.inp<double>( sect_device, s, "WindSpeedFactor");
    factors.accum_voltage  = cfg.inp<double>( sect_device, s, "AccVoltFactor");
    factors.generator_rps  = cfg.inp<double>( sect_device, s, "GenRpsFactor");
//    std::cerr << "Constructor Meteo" << std::endl;
}

Meteo::~Meteo(){
//    std::cerr << "Destructor Meteo" << std::endl;
}

void Meteo::getData( MonitorInfo& c ) {
    std::vector<unsigned char> x = askData( GET_PRIM_DATA );
    if( x.size() == 8 ) {
        c.wind_direction = fmod( factors.wind_direction*(double)*(short*)&x[0] - shifts.wind_direction + 360, 360.0 );
        c.accum_voltage = factors.accum_voltage*(double)*(short*)&x[2];
        c.temperature1  = factors.temperature1*(double)*(short*)&x[4] - shifts.temperature1;
        c.temperature2  = factors.temperature2*(double)*(short*)&x[6] - shifts.temperature2;
    }
    x = askData( GET_SECD_DATA );
    if( x.size() == 8 ) {
        c.pressure      = factors.pressure*(double)*(short*)&x[0];
        c.humidity      = factors.humidity*(double)*(short*)&x[2] - shifts.humidity;
        c.wind_speed    = factors.wind_speed*(double)*(short*)&x[4];
        c.generator_rps = factors.generator_rps*(double)*(short*)&x[6];
    }
    c.switch1 = SENS1&getSensor();
    c.switch2 = SENS2&sensor;
}

void Meteo::accumulate( std::vector<MonitorInfo>& v, MonitorInfo& acc ) {
    means( v, acc );
    maxs( v, acc );
    switchs( v, acc );
}

double Meteo::dew_point_calc( double t, double rh ) {
    static const double a = 17.27;
    static const double b = 237.7;
    double g = a*t/( b + t ) + log( 0.01*rh );
    return b*g/( a - g );
}

void Meteo::means( std::vector<MonitorInfo>& v, MonitorInfo& acc ) {
    unsigned int n = v.size();
    double base_direction = v[0].wind_direction;
    for( unsigned int i = 0; i < n; i++ ) {
        acc.temperature1 += v[i].temperature1/n;
        acc.temperature2 += v[i].temperature2/n;
        acc.humidity     += v[i].humidity/n;
        acc.pressure     += v[i].pressure/n;
        acc.accum_voltage+= v[i].accum_voltage/n;
        acc.actuat_voltage+=v[i].actuat_voltage/n;
        acc.actuat_temper+= v[i].actuat_temper/n;
        acc.generator_rps+= v[i].generator_rps/n;
        acc.wind_speed   += v[i].wind_speed/n;
        if( v[i].wind_direction - base_direction >  180.0 ) v[i].wind_direction -= 360.0;
        if( v[i].wind_direction - base_direction < -180.0 ) v[i].wind_direction += 360.0;
        acc.wind_direction += v[i].wind_direction/n;
    }
    acc.wind_direction += ( acc.wind_direction < 0.0 ) ? 360.0 : 0.0;
    acc.wind_direction -= ( acc.wind_direction > 360.0 ) ? 360.0 : 0.0;
    acc.dew_point = dew_point_calc( acc.temperature1, acc.humidity );
}

void Meteo::switchs( std::vector<MonitorInfo>& v, MonitorInfo& acc ) {
    unsigned int n = v.size();
    for( unsigned int i = 0; i < n; i++ ) {
        acc.switch1    &= v[i].switch1;
        acc.switch2    &= v[i].switch2;
    }
}

void Meteo::maxs( std::vector<MonitorInfo>& v,  MonitorInfo& acc ) {
    unsigned int n = v.size();
    for( unsigned int i = 0; i < n; i++ ) {
        acc.wind_max = ( v[i].wind_speed > acc.wind_max) ? v[i].wind_speed : acc.wind_max;
    }
}

