#ifndef __SOCKET_H_
#define __SOCKET_H_

/*
*   socket.h is a part of programm carbon
*   $Id: socket.h,v 1.1.1.1 2007/09/09 16:50:21 matwey Exp $
*   Copyright (C) 2005-2006  Matwey V. Kornilov
*
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <string>
#include <stack>
#include <sstream>
#include "singleton.h"

class IP4Addr{
private:
    std::string m_name;
    unsigned short m_port;
public:
    IP4Addr();
    IP4Addr( const std::string& name, unsigned short port );
    ~IP4Addr();
    const std::string& getName() const;
    unsigned short getPort() const;
};

namespace std{
    std::ostream& operator<<( std::ostream& stm, const IP4Addr& a );
    std::istream& operator>>( std::istream& stm, IP4Addr& a );
};

template <class SocketTraits> class BasicSocketBase;
template <class SocketTraits> class BasicClientSocket;
template <class SocketTraits> class BasicServerSocket;
template <class SocketTraits> class BasicSocketWatcher;
template <class SocketTraits> class BasicSocketError;

template <class SocketTraits> class BasicSocketBase{
    typedef BasicSocketBase < SocketTraits > SocketBase;
    typedef BasicClientSocket < SocketTraits > ClientSocket;
    typedef BasicServerSocket < SocketTraits > ServerSocket;
    typedef BasicSocketWatcher < SocketTraits > SocketWatcher;
    typedef BasicSocketError < SocketTraits > SocketError;
    friend class SocketWatcher::BaseType;
public:
    typedef typename SocketTraits::AddrType AddrType;
public:
    enum SocketState{
        Idle=1,     //!<  
        Listening,  //!<  ,   
        Connecting, //!<     
        Connected,  //!<  
    };
protected:
    virtual void onStateChanged( const typename SocketBase::SocketState s );
private:
    SocketState   m_state;
protected:
    SocketTraits* m_pImpl;
public:
    BasicSocketBase();
    virtual ~BasicSocketBase() = 0;
private:
    BasicSocketBase( const BasicSocketBase& b );
    BasicSocketBase& operator=( const BasicSocketBase& b );
protected:
    void setState( SocketState state );
public:
    SocketState getState() const;
    const AddrType getPeerName() const;
    const AddrType getSockName() const;
};

template <class SocketTraits> class BasicClientSocket: public BasicSocketBase< SocketTraits >{
protected:
    typedef BasicSocketBase < SocketTraits > SocketBase;
    typedef BasicClientSocket < SocketTraits > ClientSocket;
    typedef BasicServerSocket < SocketTraits > ServerSocket;
    typedef BasicSocketWatcher < SocketTraits > SocketWatcher;
    typedef BasicSocketError < SocketTraits > SocketError;
    friend class SocketWatcher::BaseType;
private:
    std::string m_buf;
    void read();
protected:
    virtual void onDataRecieved();
    virtual void onError( const SocketError& e );
public:
    BasicClientSocket();
    BasicClientSocket( ServerSocket* serv );
    ~BasicClientSocket();

    int hasData() const;
    void connect( const typename ClientSocket::AddrType& addr );
    void close();

    const std::string& buffer() const { return m_buf; }
    void getData(std::string& dest, std::string::size_type beg=0, std::string::size_type end=0);

    template <class T> ClientSocket& operator<<( const T& t ){
        std::ostringstream oss;
        oss << t;
        BasicSocketBase<SocketTraits>::m_pImpl->write( oss.str() );
        return *this;
    }
    template <class T> ClientSocket& operator>>( T& t ){
        std::istringstream iss( m_buf );
        iss >> t;
        m_buf.erase( 0, iss.tellg() );
        return *this;
    }
};

template <class SocketTraits> class BasicServerSocket: public BasicSocketBase< SocketTraits >{
protected:
    typedef BasicSocketBase < SocketTraits > SocketBase;
    typedef BasicClientSocket < SocketTraits > ClientSocket;
    typedef BasicServerSocket < SocketTraits > ServerSocket;
    typedef BasicSocketWatcher < SocketTraits > SocketWatcher;
    typedef BasicSocketError < SocketTraits > SocketError;
    friend class BasicClientSocket < SocketTraits >;
    friend class SocketWatcher::BaseType;
protected:
    virtual void onAccepted();
    virtual void onError( const SocketError& e );
private:
    std::stack< SocketTraits* > m_accepted;
    void accept();
    SocketTraits* popAcceptedTraits();
public:
    BasicServerSocket();
    ~BasicServerSocket();

    int hasAccepted() const;
    void listen( const typename ServerSocket::AddrType& addr, int backlog=1 );
    void close();
};

template <class SocketTraits> class BasicSocketError: public std::runtime_error {
    typedef BasicSocketBase < SocketTraits > SocketBase;
    typedef BasicClientSocket < SocketTraits > ClientSocket;
    typedef BasicServerSocket < SocketTraits > ServerSocket;
    typedef BasicSocketWatcher < SocketTraits > SocketWatcher;
    typedef BasicSocketError < SocketTraits > SocketError;
public:
    typedef typename SocketTraits::Error Error;
private:
    Error m_error;
public:
    BasicSocketError( const std::string& what, Error error = Error() );
    Error code() const;
};

template <class SocketTraits> class BasicSocketWatcher
   : public Singleton<typename SocketTraits::Watcher, MayersSingletonTraits> {
};

#include "linux.h"

typedef class BasicSocketBase   < DefinedSocketTraits > DefinedSocketBase;
typedef class BasicClientSocket < DefinedSocketTraits > DefinedClientSocket;
typedef class BasicServerSocket < DefinedSocketTraits > DefinedServerSocket;
typedef class BasicSocketWatcher< DefinedSocketTraits > DefinedSocketWatcher;
typedef class BasicSocketError  < DefinedSocketTraits > DefinedSocketError;

/*
template class BasicSocketBase< DefinedSocketTraits >;
template class BasicClientSocket< DefinedSocketTraits >;
template class BasicServerSocket< DefinedSocketTraits >;
template class BasicSocketWatcher< DefinedSocketTraits >;
template class BasicSocketError< DefinedSocketTraits >;
*/

#include "callback.h"

typedef DefinedSocketBase    SocketBase;
typedef ClientSocketWithCalls  ClientSocket;
typedef ServerSocketWithCalls  ServerSocket;
typedef DefinedSocketWatcher SocketWatcher;
typedef DefinedSocketError   SocketError;

#endif // __SOCKET_H_

